@@ -23,7 +23,7 @@ def displayhook(value: Any) -> None:
     setattr(builtins, "_", None)
     if isinstance(value, drgn.Object):
         text = value.format_(columns=shutil.get_terminal_size((0, 0)).columns)
-    elif isinstance(value, (drgn.StackTrace, drgn.Type)):
+    elif isinstance(value, (drgn.StackFrame, drgn.StackTrace, drgn.Type)):
         text = str(value)
     else:
         text = repr(value)
