@@ -1,19 +1,43 @@
 // Copyright (c) Facebook, Inc. and its affiliates.
-// SPDX-License-Identifier: GPL-3.0+
+// SPDX-License-Identifier: GPL-3.0-or-later
 
 #ifndef DRGN_PLATFORM_H
 #define DRGN_PLATFORM_H
 
-#include <elfutils/libdwfl.h>
 #include <gelf.h>
 
+#include "cfi.h"
 #include "drgn.h"
+#include "util.h"
+
+struct drgn_orc_entry;
+struct drgn_register_state;
 
 struct drgn_register {
-	const char *name;
+	const char * const *names;
+	size_t num_names;
+	drgn_register_number regno;
 	uint64_t dwarf_number;
 };
 
+struct drgn_register_layout {
+	uint32_t offset;
+	uint32_t size;
+};
+
+/* ELF section to apply relocations to. */
+struct drgn_relocating_section {
+	char *buf;
+	size_t buf_size;
+	uint64_t addr;
+	bool bswap;
+};
+
+typedef struct drgn_error *
+apply_elf_rela_fn(const struct drgn_relocating_section *relocating,
+		  uint64_t r_offset, uint32_t r_type, int64_t r_addend,
+		  uint64_t sym_value);
+
 /* Page table iterator. */
 struct pgtable_iterator {
 	struct drgn_program *prog;
@@ -22,7 +46,7 @@ struct pgtable_iterator {
 	/* Current virtual address to translate. */
 	uint64_t virt_addr;
 	/* Architecture-specific data. */
-	char arch[0];
+	char arch[];
 };
 
 /*
@@ -59,15 +83,30 @@ struct drgn_architecture_info {
 	const struct drgn_register *registers;
 	size_t num_registers;
 	const struct drgn_register *(*register_by_name)(const char *name);
+	const struct drgn_register_layout *register_layout;
+	drgn_register_number (*dwarf_regno_to_internal)(uint64_t);
+	/* CFI row containing default rules for DWARF CFI. */
+	const struct drgn_cfi_row *default_dwarf_cfi_row;
+	struct drgn_error *(*orc_to_cfi)(const struct drgn_orc_entry *,
+					 struct drgn_cfi_row **, bool *,
+					 drgn_register_number *);
+	/*
+	 * Try to unwind a stack frame if CFI wasn't found. Returns &drgn_stop
+	 * if we couldn't.
+	 */
+	struct drgn_error *(*fallback_unwind)(struct drgn_program *,
+					      struct drgn_register_state *,
+					      struct drgn_register_state **);
 	/* Given pt_regs as a value buffer object. */
-	struct drgn_error *(*pt_regs_set_initial_registers)(Dwfl_Thread *,
-							    const struct drgn_object *);
-	struct drgn_error *(*prstatus_set_initial_registers)(struct drgn_program *,
-							     Dwfl_Thread *,
+	struct drgn_error *(*pt_regs_get_initial_registers)(const struct drgn_object *,
+							    struct drgn_register_state **);
+	struct drgn_error *(*prstatus_get_initial_registers)(struct drgn_program *,
 							     const void *,
-							     size_t);
-	struct drgn_error *(*linux_kernel_set_initial_registers)(Dwfl_Thread *,
-								 const struct drgn_object *);
+							     size_t,
+							     struct drgn_register_state **);
+	struct drgn_error *(*linux_kernel_get_initial_registers)(const struct drgn_object *,
+								 struct drgn_register_state **);
+	apply_elf_rela_fn *apply_elf_rela;
 	struct drgn_error *(*linux_kernel_get_page_offset)(struct drgn_object *);
 	struct drgn_error *(*linux_kernel_get_vmemmap)(struct drgn_object *);
 	struct drgn_error *(*linux_kernel_live_direct_mapping_fallback)(struct drgn_program *,
@@ -81,15 +120,6 @@ struct drgn_architecture_info {
 	pgtable_iterator_next_fn *linux_kernel_pgtable_iterator_next;
 };
 
-static inline const struct drgn_register *
-drgn_architecture_register_by_name(const struct drgn_architecture_info *arch,
-				   const char *name)
-{
-	if (!arch->register_by_name)
-		return NULL;
-	return arch->register_by_name(name);
-}
-
 extern const struct drgn_architecture_info arch_info_unknown;
 extern const struct drgn_architecture_info arch_info_x86_64;
 extern const struct drgn_architecture_info arch_info_ppc64;
@@ -99,6 +129,34 @@ struct drgn_platform {
 	enum drgn_platform_flags flags;
 };
 
+static inline bool
+drgn_platform_is_little_endian(const struct drgn_platform *platform)
+{
+	return platform->flags & DRGN_PLATFORM_IS_LITTLE_ENDIAN;
+}
+
+static inline bool drgn_platform_bswap(const struct drgn_platform *platform)
+{
+	return drgn_platform_is_little_endian(platform) != HOST_LITTLE_ENDIAN;
+}
+
+static inline bool drgn_platform_is_64_bit(const struct drgn_platform *platform)
+{
+	return platform->flags & DRGN_PLATFORM_IS_64_BIT;
+}
+
+static inline uint8_t
+drgn_platform_address_size(const struct drgn_platform *platform)
+{
+	return drgn_platform_is_64_bit(platform) ? 8 : 4;
+}
+
+static inline uint64_t
+drgn_platform_address_mask(const struct drgn_platform *platform)
+{
+	return drgn_platform_is_64_bit(platform) ? UINT64_MAX : UINT32_MAX;
+}
+
 /**
  * Initialize a @ref drgn_platform from an architecture, word size, and
  * endianness.
