@@ -1,5 +1,5 @@
 // Copyright (c) Facebook, Inc. and its affiliates.
-// SPDX-License-Identifier: GPL-3.0+
+// SPDX-License-Identifier: GPL-3.0-or-later
 
 #ifndef DRGNPY_H
 #define DRGNPY_H
@@ -37,7 +37,7 @@
     } while (0)
 #endif
 
-#define DRGNPY_PUBLIC __attribute__((visibility("default")))
+#define DRGNPY_PUBLIC __attribute__((__visibility__("default")))
 
 #define Py_RETURN_BOOL(cond) do {	\
 	if (cond)			\
@@ -236,7 +236,6 @@ DrgnType *Program_void_type(Program *self, PyObject *args, PyObject *kwds);
 DrgnType *Program_int_type(Program *self, PyObject *args, PyObject *kwds);
 DrgnType *Program_bool_type(Program *self, PyObject *args, PyObject *kwds);
 DrgnType *Program_float_type(Program *self, PyObject *args, PyObject *kwds);
-DrgnType *Program_complex_type(Program *self, PyObject *args, PyObject *kwds);
 DrgnType *Program_struct_type(Program *self, PyObject *args, PyObject *kwds);
 DrgnType *Program_union_type(Program *self, PyObject *args, PyObject *kwds);
 DrgnType *Program_class_type(Program *self, PyObject *args, PyObject *kwds);
@@ -249,14 +248,6 @@ DrgnType *Program_function_type(Program *self, PyObject *args, PyObject *kwds);
 int append_string(PyObject *parts, const char *s);
 int append_format(PyObject *parts, const char *format, ...);
 PyObject *join_strings(PyObject *parts);
-PyObject *byteorder_string(bool little_endian);
-
-struct byteorder_arg {
-	bool allow_none;
-	bool is_none;
-	enum drgn_byte_order value;
-};
-int byteorder_converter(PyObject *o, void *p);
 
 struct index_arg {
 	bool allow_none;
